from rest_framework.filters import BaseFilterBackend
from rest_framework_guardian.filters import ObjectPermissionsFilter
from djangoldp.utils import is_anonymous_user


class LDPPermissionsFilterBackend(ObjectPermissionsFilter):
    """
    Default FilterBackend for LDPPermissions. If user does not have model-level permissions, filters by
    Django-Guardian's get_objects_for_user
    """

    shortcut_kwargs = {
        'accept_global_perms': False,
        'with_superuser': True
    }

    def filter_queryset(self, request, queryset, view):
        from djangoldp.models import Model
        from djangoldp.permissions import LDPPermissions, ModelConfiguredPermissions

        # compares the requirement for GET, with what the user has on the MODEL
        ldp_permissions = LDPPermissions()
        if ldp_permissions.has_container_permission(request, view):
            return queryset

        if not is_anonymous_user(request.user):
            # those objects I have by grace of group or object
            # first figure out if the superuser has special permissions (important to the implementation in superclass)
            perms_class = ModelConfiguredPermissions()
            anon_perms, auth_perms, owner_perms, superuser_perms = perms_class.get_permission_settings(view.model)
            self.shortcut_kwargs['with_superuser'] = 'view' in superuser_perms

            object_perms = super().filter_queryset(request, queryset, view)

            # those objects I have by grace of being owner
            if Model.get_meta(view.model, 'owner_field', None) is not None:
                if 'view' in owner_perms:
                    owned_objects = [q.pk for q in queryset if Model.is_owner(view.model, request.user, q)]
                    return object_perms | queryset.filter(pk__in=owned_objects)
            return object_perms

        # user is anonymous without anonymous permissions
        return view.model.objects.none()


class LocalObjectFilterBackend(BaseFilterBackend):
    """
    Filter which removes external objects (federated backlinks) from the queryset
    For querysets which should only include local objects
    """
    def filter_queryset(self, request, queryset, view):
        from djangoldp.models import Model

        internal_ids = [x.pk for x in queryset if not Model.is_external(x)]
        return queryset.filter(pk__in=internal_ids)


class LocalObjectOnContainerPathBackend(LocalObjectFilterBackend):
    """
    Override of LocalObjectFilterBackend which removes external objects when the view requested
    is the model container path
    """
    def filter_queryset(self, request, queryset, view):
        from djangoldp.models import Model

        if issubclass(view.model, Model) and request.path_info == view.model.get_container_path():
            return super(LocalObjectOnContainerPathBackend, self).filter_queryset(request, queryset, view)
        return queryset



from Cryptodome.PublicKey import RSA
from django.core.management.base import BaseCommand
from djangoldp_crypto.models import RSAKey


class Command(BaseCommand):
    help = 'Randomly generate a new RSA key for the DjangoLDP server'

    def handle(self, *args, **options):
        try:
            key = RSA.generate(2048)
            rsakey = RSAKey(priv_key=key.exportKey('PEM').decode('utf8'))
            rsakey.save()
            self.stdout.write('RSA key successfully created')
            self.stdout.write(u'Private key: \n{0}'.format(rsakey.priv_key))
            self.stdout.write(u'Public key: \n{0}'.format(rsakey.pub_key))
            self.stdout.write(u'Key ID: \n{0}'.format(rsakey.kid))
        except Exception as e:
            self.stdout.write('Something goes wrong: {0}'.format(e))



from django.conf import settings
from django.test import TestCase

class TestSettings(TestCase):

    def test_inexistent_returns_default(self):
        """Assert a inexistent key returns the provided default value."""
        assert getattr(settings, 'INEXISTENT', 'something') == 'something'

    def test_only_in_core_config(self):
        """Asserts values defined only in core config."""
        assert settings.DEBUG == False

    def test_only_in_package(self):
        """Asserts default settings defined in the package."""
        assert settings.MYPACKAGEVAR == "ok"

    def test_only_in_user_config(self):
        """Asserts LDP packages are loaded from YAML file."""
        assert settings.DJANGOLDP_PACKAGES == ['djangoldp.tests']

    def test_overrided_core_by_package_config(self):
        assert settings.USE_I18N == False

    def test_overrided_package_by_user_config(self):
        assert settings.USE_TZ == False

    def test_overrided_core_by_user_config(self):
        """Asserts values overrided from user configuration."""
        assert settings.EMAIL_HOST == 'somewhere'

    def test_installed_apps_resolution(self):
        """Asserts LDP packages are referenced along with default installed apps."""
        assert settings.INSTALLED_APPS == [
            'djangoldp.tests',
            'djangoldp.tests.dummy.apps.DummyConfig',
            'django.contrib.admin',
            'django.contrib.auth',
            'django.contrib.contenttypes',
            'django.contrib.sessions',
            'django.contrib.messages',
            'django.contrib.staticfiles',
            'djangoldp',
            'guardian',
        ]

    def test_reference_middleware(self):
        """Asserts middlewares added in packages are added to the settings."""
        assert 'djangoldp.tests.dummy.middleware.DummyMiddleware' in settings.MIDDLEWARE

    def test_extra_module(self):
        #FIXME
        pass



'''
  DjangoLDP Check Integrity Command & Importer
  Usage `./manage.py check_integrity --help`

  This command does import every `check_integrity.py` of every DJANGOLDP_PACKAGES
  Allowed methods:
    `add_arguments(parser)`: Allows to create new arguments to this command
    `check_integrity(options)`: Do your own checks on the integrity
  
  Examples on `djangoldp/check_integrity.py`
'''

from django.core.management.base import BaseCommand
from django.conf import settings
from importlib import import_module
import requests

class Command(BaseCommand):
  help = "Check the datas integrity"

  def add_arguments(self, parser):

    import_module('djangoldp.check_integrity').add_arguments(parser)

    for package in settings.DJANGOLDP_PACKAGES:
      try:
          import_module('{}.check_integrity'.format(package)).add_arguments(parser)
      except:
          pass

  def handle(self, *args, **options):

    import_module('djangoldp.check_integrity').check_integrity(options)

    for package in settings.DJANGOLDP_PACKAGES:
      try:
          import_module('{}.check_integrity'.format(package)).check_integrity(options)
      except:
          pass

    exit(0)



'''
  DjangoLDP Check Integrity
  Usage `./manage.py check_integrity --help`
'''
from django.apps import apps
from django.conf import settings
from djangoldp.models import LDPSource
from urllib.parse import urlparse
import requests

# Helper command for argument type checking
def is_string(target):
  if isinstance(target, str):
    return target
  return False

# Helper command to check status code of a target
def is_alive(target, status_code = 200):
  return requests.get(target).status_code == status_code

# Add argument to the `check_integrity` command
def add_arguments(parser):
  parser.add_argument(
    "--ignore",
    action="store",
    default=False,
    type=is_string,
    help="Ignore any server, comma separated",
  )
  parser.add_argument(
    "--ignore-faulted",
    default=False,
    nargs="?",
    const=True,
    help="Ignore eventual faulted",
  )
  parser.add_argument(
    "--ignore-404",
    default=False,
    nargs="?",
    const=True,
    help="Ignore eventual 404",
  )
  parser.add_argument(
    "--fix-faulted-resources",
    default=False,
    nargs="?",
    const=True,
    help="Fix faulted resources",
  )
  parser.add_argument(
    "--fix-404-resources",
    default=False,
    nargs="?",
    const=True,
    help="Fix 404 resources",
  )
  parser.add_argument(
    "--fix-offline-servers",
    default=False,
    nargs="?",
    const=True,
    help="Remove resources from offline servers",
  )

# Define our own checks
def check_integrity(options):

  models = apps.get_models()

  ignored = set()
  if(options["ignore"]):
    for target in options["ignore"].split(","):
      ignored.add(urlparse(target).netloc)

  if(len(ignored) > 0):
    print("Ignoring servers:")
    for server in ignored:
      print("- "+server)

  resources = set()
  resources_map = dict()
  base_urls = set()

  for model in models:
    for obj in model.objects.all():
      if hasattr(obj, "urlid"):
        if(obj.urlid):
          if(not obj.urlid.startswith(settings.BASE_URL)):
            url = urlparse(obj.urlid).netloc
            if(url not in ignored):
              resources.add(obj.urlid)
              resources_map[obj.urlid] = obj
              base_urls.add(url)

  if(len(base_urls) > 0):
    print("Found "+str(len(resources_map))+" distant resources on "+str(len(models))+" models")
    print("Servers that I have backlinks to:")
    for server in base_urls:
      print("- "+server)
  else:
    print("I don't have any backlink")

  source_urls = set()
  for source in LDPSource.objects.all():
    source_urls.add(urlparse(source.urlid).netloc)

  if(len(source_urls) > 0):
    print("Servers that I'm allowed to get federated to:")
    for server in source_urls:
      print("- "+server)
  else:
    print("I'm not federated")

  difference_servers = base_urls.difference(source_urls)
  if(len(difference_servers) > 0):
    print("Servers that I should not get aware of:")
    for server in difference_servers:
      print("- "+server)

    # Handle faulted resources
    if(not options["ignore_faulted"]):
      faulted_resources = set()
      for server in difference_servers:
        for resource in resources:
          if(urlparse(resource).netloc in server):
            faulted_resources.add(resource)

      if(len(faulted_resources) > 0):
        print("Resources in fault:")
        for resource in faulted_resources:
          print("- "+resource)
      else:
        print("No resource are in fault")

      if(options["fix_faulted_resources"]):
        for resource in faulted_resources:
          try:
            resources_map[resource].delete()
          except:
            pass
        print("Fixed faulted resources")
      else:
        print("Fix them with `./manage.py check_integrity --fix-faulted-resources`")
  else:
    print("I accept datas for every of those servers")

  # Handle 404 resources
  if(not options["ignore_404"]):
    resources_404 = set()
    resources_servers_offline = set()
    for resource in resources:
      try:
        if(is_alive(resource, 404)):
          resources_404.add(resource)
      except:
        resources_servers_offline.add(resource)

    if(len(resources_404) > 0):
      print("Faulted resources, 404:")
      for resource in resources_404:
        print("- "+resource)
      if(options["fix_404_resources"]):
        for resource in resources_404:
          try:
            resources_map[resource].delete()
          except:
            pass
        print("Fixed 404 resources")
      else:
        print("Fix them with `./manage.py check_integrity --fix-404-resources`")

    if(len(resources_servers_offline) > 0):
      print("Faulted resources, servers offline:")
      for resource in resources_servers_offline:
        print("- "+resource)
      if(options["fix_offline_servers"]):
        for resource in resources_servers_offline:
          try:
            resources_map[resource].delete()
          except:
            pass
        print("Fixed resources on offline servers")
      else:
        print("Fix them with `./manage.py check_integrity --fix-offline-servers`")

    else:
      print("No 404 in known resources")